@@ -7,7 +7,7 @@ use Carp ();
 use IO::Handle ();
 use Exporter 'import';
 
-our $VERSION = '0.005';
+our $VERSION = '0.006';
 
 our @EXPORT = 'cgi';
 
@@ -194,17 +194,8 @@ sub server_software   { defined $ENV{SERVER_SOFTWARE} ? $ENV{SERVER_SOFTWARE} :
 *path = \&path_info;
 *query = \&query_string;
 
-sub query_pairs { [@{$_[0]->_query_params->{ordered}}] }
-sub query_params {
-  my ($self) = @_;
-  my %params;
-  my $keyed = $self->_query_params->{keyed};
-  foreach my $key (keys %$keyed) {
-    my @values = @{$keyed->{$key}};
-    $params{$key} = @values > 1 ? \@values : $values[0];
-  }
-  return \%params;
-}
+sub query_params      { [map { [@$_] } @{$_[0]->_query_params->{ordered}}] }
+sub query_param_names { [keys %{$_[0]->_query_params->{keyed}}] }
 sub query_param       { my $p = $_[0]->_query_params->{keyed}; exists $p->{$_[1]} ? $p->{$_[1]}[-1] : undef }
 sub query_param_array { my $p = $_[0]->_query_params->{keyed}; exists $p->{$_[1]} ? [@{$p->{$_[1]}}] : [] }
 
@@ -224,10 +215,7 @@ sub _query_params {
   return $self->{query_params};
 }
 
-sub headers { +{%{$_[0]->_headers}} }
-sub header { (my $name = $_[1]) =~ tr/-/_/; $ENV{"HTTP_\U$name"} }
-
-sub _headers {
+sub headers {
   my ($self) = @_;
   unless (exists $self->{request_headers}) {
     my %headers;
@@ -239,23 +227,30 @@ sub _headers {
     }
     $self->{request_headers} = \%headers;
   }
-  return $self->{request_headers};
+  return {%{$self->{request_headers}}};
 }
 
-sub cookies { +{%{$_[0]->_cookies}} }
-sub cookie { $_[0]->_cookies->{$_[1]} }
+sub header { (my $name = $_[1]) =~ tr/-/_/; $ENV{"HTTP_\U$name"} }
+
+sub cookies      { [map { [@$_] } @{$_[0]->_cookies->{ordered}}] }
+sub cookie_names { [keys %{$_[0]->_cookies->{keyed}}] }
+sub cookie       { my $c = $_[0]->_cookies->{keyed}; exists $c->{$_[1]} ? $c->{$_[1]}[-1] : undef }
+sub cookie_array { my $c = $_[0]->_cookies->{keyed}; exists $c->{$_[1]} ? [@{$c->{$_[1]}}] : [] }
 
 sub _cookies {
   my ($self) = @_;
   unless (exists $self->{request_cookies}) {
-    $self->{request_cookies} = {};
+    my (@ordered, %keyed);
     if (defined $ENV{HTTP_COOKIE}) {
       foreach my $pair (split /\s*;\s*/, $ENV{HTTP_COOKIE}) {
         next unless length $pair;
         my ($name, $value) = split /=/, $pair, 2;
-        $self->{request_cookies}{$name} = $value if defined $value;
+        next unless defined $value;
+        push @ordered, [$name, $value];
+        push @{$keyed{$name}}, $value;
       }
     }
+    $self->{request_cookies} = {ordered => \@ordered, keyed => \%keyed};
   }
   return $self->{request_cookies};
 }
@@ -286,17 +281,8 @@ sub body {
   return $self->{content};
 }
 
-sub body_pairs { [@{$_[0]->_body_params->{ordered}}] }
-sub body_params {
-  my ($self) = @_;
-  my %params;
-  my $keyed = $self->_body_params->{keyed};
-  foreach my $key (keys %$keyed) {
-    my @values = @{$keyed->{$key}};
-    $params{$key} = @values > 1 ? \@values : $values[0];
-  }
-  return \%params;
-}
+sub body_params      { [map { [@$_] } @{$_[0]->_body_params->{ordered}}] }
+sub body_param_names { [keys %{$_[0]->_body_params->{keyed}}] }
 sub body_param       { my $p = $_[0]->_body_params->{keyed}; exists $p->{$_[1]} ? $p->{$_[1]}[-1] : undef }
 sub body_param_array { my $p = $_[0]->_body_params->{keyed}; exists $p->{$_[1]} ? [@{$p->{$_[1]}}] : [] }
 
@@ -330,40 +316,12 @@ sub body_json {
   return $self->{body_json};
 }
 
-sub add_response_header {
-  my ($self, $name, $value) = @_;
+sub set_nph {
+  my ($self, $value) = @_;
   if ($self->{headers_rendered}) {
-    Carp::carp "Attempted to add HTTP response header '$name' but headers have already been rendered";
+    Carp::carp "Attempted to set NPH response mode but headers have already been rendered";
   } else {
-    Carp::croak "Newline characters not allowed in HTTP response header '$name'" if $value =~ tr/\r\n//;
-    push @{$self->{response_headers}}, [$name, $value];
-  }
-  return $self;
-}
-
-my %COOKIE_ATTR_VALUE = (expires => 1, domain => 1, path => 1, secure => 0, httponly => 0, samesite => 1, 'max-age' => 1);
-sub add_response_cookie {
-  my ($self, $name, $value, @attrs) = @_;
-  if ($self->{headers_rendered}) {
-    Carp::carp "Attempted to add HTTP response cookie '$name' but headers have already been rendered";
-  } else {
-    my $cookie_str = "$name=$value";
-    my $i = 0;
-    while ($i <= $#attrs) {
-      my ($key, $val) = @attrs[$i, $i+1];
-      my $has_value = $COOKIE_ATTR_VALUE{lc $key};
-      if (!defined $has_value) {
-        Carp::carp "Attempted to set unknown cookie attribute '$key'";
-      } elsif ($has_value) {
-        $cookie_str .= "; $key=$val" if defined $val;
-      } else {
-        $cookie_str .= "; $key" if $val;
-      }
-    } continue {
-      $i += 2;
-    }
-    Carp::croak "Newline characters not allowed in HTTP response cookie '$name'" if $cookie_str =~ tr/\r\n//;
-    push @{$self->{response_headers}}, ['Set-Cookie', $cookie_str];
+    $self->{nph} = $value;
   }
   return $self;
 }
@@ -401,12 +359,40 @@ sub set_response_charset {
   return $self;
 }
 
-sub set_nph {
-  my ($self, $value) = @_;
+sub add_response_header {
+  my ($self, $name, $value) = @_;
   if ($self->{headers_rendered}) {
-    Carp::carp "Attempted to set NPH response mode but headers have already been rendered";
+    Carp::carp "Attempted to add HTTP response header '$name' but headers have already been rendered";
   } else {
-    $self->{nph} = $value;
+    Carp::croak "Newline characters not allowed in HTTP response header '$name'" if $value =~ tr/\r\n//;
+    push @{$self->{response_headers}}, [$name, $value];
+  }
+  return $self;
+}
+
+my %COOKIE_ATTR_VALUE = (expires => 1, domain => 1, path => 1, secure => 0, httponly => 0, samesite => 1, 'max-age' => 1);
+sub add_response_cookie {
+  my ($self, $name, $value, @attrs) = @_;
+  if ($self->{headers_rendered}) {
+    Carp::carp "Attempted to add HTTP response cookie '$name' but headers have already been rendered";
+  } else {
+    my $cookie_str = "$name=$value";
+    my $i = 0;
+    while ($i <= $#attrs) {
+      my ($key, $val) = @attrs[$i, $i+1];
+      my $has_value = $COOKIE_ATTR_VALUE{lc $key};
+      if (!defined $has_value) {
+        Carp::carp "Attempted to set unknown cookie attribute '$key'";
+      } elsif ($has_value) {
+        $cookie_str .= "; $key=$val" if defined $val;
+      } else {
+        $cookie_str .= "; $key" if $val;
+      }
+    } continue {
+      $i += 2;
+    }
+    Carp::croak "Newline characters not allowed in HTTP response cookie '$name'" if $cookie_str =~ tr/\r\n//;
+    push @{$self->{response_headers}}, ['Set-Cookie', $cookie_str];
   }
   return $self;
 }
@@ -471,8 +457,12 @@ sub render {
   if ($type eq 'json') {
     $out_fh->printflush($self->_json->encode($data));
   } elsif ($type eq 'html' or $type eq 'xml' or $type eq 'text') {
-    require Encode;
-    $out_fh->printflush(Encode::encode($charset, "$data"));
+    if (uc $charset eq 'UTF-8' and do { local $@; eval { require Unicode::UTF8; 1 } }) {
+      $out_fh->printflush(Unicode::UTF8::encode_utf8($data));
+    } else {
+      require Encode;
+      $out_fh->printflush(Encode::encode($charset, "$data"));
+    }
   } elsif ($type eq 'data') {
     $out_fh->printflush($data);
   }
@@ -481,8 +471,7 @@ sub render {
 sub _json {
   my ($self) = @_;
   unless (exists $self->{json}) {
-    local $@;
-    if (eval { require Cpanel::JSON::XS; Cpanel::JSON::XS->VERSION('4.09'); 1 }) {
+    if (do { local $@; eval { require Cpanel::JSON::XS; Cpanel::JSON::XS->VERSION('4.09'); 1 } }) {
       $self->{json} = Cpanel::JSON::XS->new->allow_dupkeys->stringify_infnan;
     } else {
       require JSON::PP;
